import numpy as np

from lda11 import LDA

from .conftest import Docs


def test_lda(docs_gen: Docs) -> None:
    (X1, _), true_theta = docs_gen.gen_doc(1000)
    lda = LDA(
        2,
        n_iter=50,
        optimize_interval=1,
        optimize_burn_in=25,
        use_cgs_p=True,
        n_workers=4,
    )
    lda.fit(X1)
    phi1 = lda.phi

    # determin which is TOPIC1

    lang1_topic1_strong_index = np.where(docs_gen.languages[0].topic_1 > 0.1)[0]
    lang1_topic2_strong_index = np.where(docs_gen.languages[0].topic_1 < 0.1)[0]
    if (
        phi1[lang1_topic1_strong_index, 0].mean()
        > phi1[lang1_topic2_strong_index, 0].mean()
    ):
        topic1_index = 0
        topic2_index = 1
    else:
        topic1_index = 1
        topic2_index = 0
    for i in lang1_topic1_strong_index:
        for j in lang1_topic2_strong_index:
            assert phi1[i, topic1_index] > phi1[j, topic1_index]
            assert phi1[i, topic2_index] < phi1[j, topic2_index]



import random
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
from numpy import typing as npt
from scipy import sparse as sps
from tqdm import tqdm
from typing_extensions import Literal

from ._lda import LDATrainer
from ._lda import Predictor as CorePredictor
from ._lda import learn_dirichlet, learn_dirichlet_symmetric, log_likelihood_doc_topic

RealType = np.float64

IntegerType = np.int32
IndexType = np.uint64


ValidXType = Union[sps.spmatrix, npt.NDArray[np.int32], npt.NDArray[np.int64]]
PriorType = Union[np.ndarray, float, None]


class LDAInput(NamedTuple):
    counts: np.ndarray
    dix: np.ndarray
    wis: np.ndarray


def number_to_array(
    n_components: int,
    default: float,
    arg_: Union[float, None, np.ndarray] = None,
    ensure_symmetry: bool = False,
) -> npt.NDArray[np.float64]:
    if arg_ is None or isinstance(arg_, float):
        value_ = default if arg_ is None else float(arg_)
        return np.ones(n_components, dtype=RealType) * value_
    if isinstance(arg_, np.ndarray):
        assert arg_.shape[0] == n_components
        if ensure_symmetry and np.unique(arg_).shape[0] > 1:
            raise ValueError("Symmetric array required.")
        return arg_.astype(RealType)
    raise ValueError("Number of ndarray is required.")


def check_array(X: ValidXType) -> LDAInput:
    assert X.dtype == np.int32 or X.dtype == np.int64
    if isinstance(X, np.ndarray):
        assert len(X.shape) == 2
        dix, wix = X.nonzero()
        counts: np.ndarray = X[dix, wix]
    elif sps.issparse(X):
        # if X is either types of, scipy.sparse X has this attribute.
        X = sps.csr_matrix(X)
        X.sort_indices()
        dix, wix = X.nonzero()
        counts = X.data.astype(np.int32)
    else:
        raise ValueError("The input must be either np.ndarray or sparse array.")
    return LDAInput(
        counts.astype(IntegerType), dix.astype(IndexType), wix.astype(IndexType)
    )


def bow_row_to_counts(X: ValidXType, i: int) -> Tuple[np.ndarray, np.ndarray]:
    wix: np.ndarray
    if isinstance(X, np.ndarray):
        assert len(X.shape) == 2
        assert X.dtype == np.int32 or X.dtype == np.int64
        (wix,) = X[i].nonzero()
        counts: np.ndarray = X[i, wix]
    else:
        _, wix = X[i].nonzero()
        counts = X[i, wix].toarray().ravel()

    return counts.astype(IntegerType), wix.astype(IndexType)


def to_valid_csr(X: ValidXType) -> sps.csr_matrix:
    result = sps.csr_matrix(X)
    result.data = result.data.astype(IntegerType)
    return result


class LDAPredictorMixin:
    topic_word_priors_: Optional[List[np.ndarray]]
    predictor: Optional[CorePredictor]

    def transform(
        self,
        *Xs: Union[ValidXType, None],
        n_iter: int = 100,
        random_seed: int = 0,
        mode: Literal["gibbs", "mf"] = "gibbs",
        mf_tolerance: float = 1e-10,
        gibbs_burn_in: int = 10,
        use_cgs_p: bool = True,
        n_workers: int = 1
    ) -> npt.NDArray[RealType]:
        assert self.topic_word_priors_ is not None
        assert self.predictor is not None
        shapes = set({int(X.shape[0]) for X in Xs if X is not None})
        if len(shapes) != 1:
            raise ValueError("Got different shape for Xs.")
        shape = list(shapes)[0]

        Xs_csr: List[sps.csr_matrix] = []
        for i, X in enumerate(Xs):
            if X is None:
                Xs_csr.append(
                    sps.csr_matrix(
                        (shape, self.topic_word_priors_[i].shape[0]),
                        dtype=IntegerType,
                    )
                )
            else:
                Xs_csr.append(to_valid_csr(X))

        if mode == "gibbs":
            return self.predictor.predict_gibbs_batch(
                Xs_csr, n_iter, gibbs_burn_in, random_seed, use_cgs_p, n_workers
            )
        elif mode == "mf":
            return self.predictor.predict_mf_batch(
                Xs_csr, n_iter, mf_tolerance, n_workers
            )
        else:
            raise ValueError('"mode" argument must be either "gibbs" for "mf".')

    def word_topic_assignment(
        self,
        *Xs: Union[ValidXType, None],
        n_iter: int = 100,
        random_seed: int = 0,
        gibbs_burn_in: int = 10,
        use_cgs_p: bool = True
    ) -> List[Tuple[np.ndarray, List[Dict[int, np.ndarray]]]]:
        assert self.topic_word_priors_ is not None
        assert self.predictor is not None
        n_domains = len(Xs)
        shapes = set({X.shape[0] for X in Xs if X is not None})
        if len(shapes) != 1:
            raise ValueError("Got different shape for Xs.")

        shape = list(shapes)[0]
        Xs_csr = []
        for i, X in enumerate(Xs):
            if X is None:
                Xs_csr.append(
                    sps.csr_matrix(
                        (shape, self.topic_word_priors_[i].shape[0]), dtype=IntegerType
                    )
                )
        results = []
        for i in range(shape):
            counts = []
            wixs = []
            for n in range(n_domains):
                count, wix = bow_row_to_counts(Xs[n], i)
                counts.append(count)
                wixs.append(wix)

            results.append(
                self.predictor.predict_gibbs_with_word_assignment(
                    wixs, counts, n_iter, gibbs_burn_in, random_seed, use_cgs_p
                )
            )
        return results

    @property
    def phis(self) -> List[npt.NDArray[RealType]]:
        assert self.predictor is not None
        return self.predictor.phis


class LDABase(LDAPredictorMixin):
    def __init__(
        self,
        n_components: int = 100,
        doc_topic_prior: PriorType = None,
        n_iter: int = 1000,
        optimize_interval: Optional[int] = None,
        optimize_burn_in: Optional[int] = None,
        n_workers: int = 1,
        use_cgs_p: bool = True,
        is_phi_symmetric: bool = True,
        random_seed: Optional[int] = 0,
    ):
        n_components = int(n_components)
        assert n_iter >= 1
        assert n_components >= 1
        self.n_components = n_components

        self.doc_topic_prior = number_to_array(
            self.n_components, 1 / float(self.n_components), doc_topic_prior
        )
        self.topic_word_priors_ = None
        self.is_phi_symmetric = is_phi_symmetric
        self.n_vocabs: Optional[List[int]] = None
        self.docstate_ = None
        self.components_: Optional[int] = None
        self.n_modals: Optional[int] = None

        self.predictor: Optional[CorePredictor] = None
        self.use_cgs_p: bool = use_cgs_p

        self.n_iter = n_iter
        self.optimize_interval = optimize_interval
        if optimize_interval is not None:
            if optimize_burn_in is None:
                optimize_burn_in = n_iter // 2
            else:
                optimize_burn_in = optimize_burn_in
        self.optimize_burn_in = optimize_burn_in

        self.n_workers = n_workers
        if random_seed is None:
            random_seed = random.randint(-(2 ** 31), 2 ** 31 - 1)
        self.random_seed = random_seed

    def _fit(self, *Xs: ValidXType, ll_freq: int = 10) -> npt.NDArray[IntegerType]:
        """
        Xs should be a list of contents.
        All entries must have the same shape[0].
        """

        self.modality = len(Xs)

        topic_word_priors_canonical: List[npt.NDArray[RealType]] = []

        doc_tuples: List[LDAInput] = []

        n_rows: Optional[int] = None
        for X in Xs:
            doc_tuples.append(check_array(X))
            if n_rows is None:
                n_rows = X.shape[0]
            else:
                assert n_rows == X.shape[0]
            topic_word_priors_canonical.append(
                number_to_array(
                    X.shape[1],
                    1 / float(self.n_components),
                    ensure_symmetry=True,
                )
            )
        if n_rows is None:
            raise ValueError("At least one doc-term matrix must be given.")

        doc_topic: npt.NDArray[IntegerType] = np.zeros(
            (n_rows, self.n_components), dtype=IntegerType
        )

        topic_counts: npt.NDArray[IntegerType] = np.zeros(
            self.n_components, dtype=IntegerType
        )

        word_topics: List[npt.NDArray[IntegerType]] = [
            np.zeros((X.shape[1], self.n_components), dtype=IntegerType) for X in Xs
        ]

        docstates: List[LDATrainer] = []
        for (count, dix, wix), word_topic in zip(doc_tuples, word_topics):
            docstate = LDATrainer(
                self.doc_topic_prior,
                count,
                dix,
                wix,
                self.n_components,
                self.random_seed,
                self.n_workers,
            )
            docstates.append(docstate)
            docstate.initialize(word_topic, doc_topic, topic_counts)
        doc_length: npt.NDArray[IntegerType] = doc_topic.sum(axis=1).astype(IntegerType)

        ll = log_likelihood_doc_topic(self.doc_topic_prior, doc_topic, doc_length)
        for topic_word_prior, word_topic, docstate in zip(
            topic_word_priors_canonical, word_topics, docstates
        ):
            ll += docstate.log_likelihood(topic_word_prior, word_topic)

        with tqdm(range(self.n_iter)) as pbar:
            pbar.set_description("Log Likelihood = {0:.2f}".format(ll))
            for i in pbar:
                for topic_word_prior, word_topic, docstate in zip(
                    topic_word_priors_canonical, word_topics, docstates
                ):
                    docstate.iterate_gibbs(
                        topic_word_prior, doc_topic, word_topic, topic_counts
                    )
                if (i + 1) % ll_freq == 0:
                    ll = log_likelihood_doc_topic(
                        self.doc_topic_prior, doc_topic, doc_length
                    )

                    for topic_word_prior, word_topic, docstate in zip(
                        topic_word_priors_canonical, word_topics, docstates
                    ):
                        ll += docstate.log_likelihood(topic_word_prior, word_topic)
                    pbar.set_description("Log Likelihood = {0:.2f}".format(ll))

                if (
                    (self.optimize_interval is not None)
                    and (i >= self.optimize_burn_in)
                    and (i % self.optimize_interval == 0)
                ):
                    doc_topic_prior_new = learn_dirichlet(
                        doc_topic,
                        self.doc_topic_prior,
                        0.1,
                        1 / float(self.n_components),
                        100,
                    )
                    for topic_word_prior, word_topic, docstate in zip(
                        topic_word_priors_canonical, word_topics, docstates
                    ):
                        if self.is_phi_symmetric:
                            topic_word_prior_new = np.ones_like(
                                topic_word_prior
                            ) * learn_dirichlet_symmetric(
                                word_topic.transpose().copy(),
                                topic_word_prior.mean(),
                                0.1,
                                1 / float(self.n_components),
                                100,
                            )
                        else:
                            topic_word_prior_new = learn_dirichlet(
                                word_topic.transpose().copy(),
                                topic_word_prior,
                                0.1,
                                1 / float(self.n_components),
                                100,
                            )
                        topic_word_prior[:] = topic_word_prior_new
                        self.doc_topic_prior = doc_topic_prior_new
                        docstate.set_doc_topic_prior(doc_topic_prior_new)
        self.topic_word_priors_ = topic_word_priors_canonical

        predictor = CorePredictor(
            self.n_components, self.doc_topic_prior, self.random_seed
        )

        for i, (twp, wt, docstate) in enumerate(
            zip(self.topic_word_priors_, word_topics, docstates)
        ):
            if self.use_cgs_p:
                phi = docstate.obtain_phi(twp, doc_topic, wt, topic_counts)
            else:
                phi = wt + twp[:, np.newaxis]
                phi /= phi.sum(axis=0)[np.newaxis, :]
                phi = phi.transpose()
            predictor.add_beta(phi.transpose())

        self.predictor = predictor

        return doc_topic


class MultilingualLDA(LDABase):
    def fit(self, *X: ValidXType, ll_freq: int = 10) -> "MultilingualLDA":
        self._fit(*X, ll_freq=ll_freq)
        return self


class LDA(LDABase):
    def fit(self, X: ValidXType, ll_freq: int = 10) -> "LDA":
        self._fit(X, ll_freq=ll_freq)
        return self

    @property
    def phi(self) -> np.ndarray:
        assert self.predictor is not None
        return self.predictor.phis[0]



import numpy as np
import pytest

from .language import Docs, Language


@pytest.fixture
def docs_gen() -> Docs:
    language_1 = Language(
        np.asfarray([1, 1, 1, 0.01, 0.01, 0.01]),
        np.asfarray([0.01, 0.01, 0.01, 1, 1, 1]),
    )
    language_2 = Language(
        np.asfarray([1, 0.01, 1, 0.01]), np.asfarray([0.01, 1, 0.01, 1])
    )
    return Docs([language_1, language_2])



from pkg_resources import DistributionNotFound, get_distribution  # type: ignore

from .labelled_lda import LabelledLDA
from .lda import LDA, MultilingualLDA
from .util import rowwise_train_test_split

try:
    __version__ = get_distribution("lda11").version
except DistributionNotFound:  # pragma: no cover
    __version__ = "unknown"

__all__ = [
    "__version__",
    "LDA",
    "LabelledLDA",
    "MultilingualLDA",
    "rowwise_train_test_split",
]



# This example requires scikit-leran.
from time import time

import numpy as np
from scipy import sparse as sps
from sklearn.datasets import fetch_20newsgroups
from sklearn.decomposition import LatentDirichletAllocation as LDA_vb
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.model_selection import train_test_split

from lda11 import LDA as LDA_cgs_p
from lda11.util import rowwise_train_test_split

N_TOPICS = 16
print("reading data...")
dataset = fetch_20newsgroups(shuffle=False, remove=("headers", "footers", "quotes"))
data_samples = dataset.data
train_docs, test_docs = train_test_split(data_samples, random_state=42)

print("priparing Count Vectorizer")
tf_vectorizer = CountVectorizer(max_df=1.0, stop_words="english")

X_train = tf_vectorizer.fit_transform(train_docs)
X_test = tf_vectorizer.transform(test_docs)

feature_names = tf_vectorizer.get_feature_names()

tf_vectorizer.get_stop_words()

print("Splitting test documents...")
X_test_train, X_test_test = rowwise_train_test_split(X_test, random_seed=114514)


print("Start fitting sk-learn model...")
start = time()
vb_model = LDA_vb(n_components=N_TOPICS)
vb_model.fit(X_train)

phi_vb = vb_model.components_ / vb_model.components_.sum(axis=1)[:, np.newaxis]
end = time()
print("done in {:.2f} seconds".format((end - start)))

print("Start fitting our lda model...")
start = time()
cgs_p_model = LDA_cgs_p(n_components=N_TOPICS, n_iter=500)
cgs_p_model.fit(X_train)
phi_cgs_p = cgs_p_model.phi.transpose()
end = time()
print("done in {:.2f} seconds".format((end - start)))

print("Start fitting paralleized CGS sampler with hyper-parameter optimization...")
start = time()
parallel_cgs_model = LDA_cgs_p(
    n_components=N_TOPICS, n_iter=500, n_workers=2, optimize_interval=50
)
parallel_cgs_model.fit(X_train)
phi_parallel_cgs = parallel_cgs_model.phi.transpose()
end = time()
print("done in {:.2f} seconds".format((end - start)))


def test_perplexity(model, phi, **kwargs):
    theta = model.transform(X_test_train, **kwargs)
    log_ps = np.log(theta.dot(phi))
    coo = X_test_test.tocoo()
    # perplexity
    return np.exp(-(log_ps[coo.row, coo.col] * coo.data).sum() / coo.data.sum())


print("Start testing vb model")
start = time()
ll_vb = test_perplexity(vb_model, phi_vb)
end = time()
print("Done in {:.2f} seconds, test perplexity = {:.2f}".format(end - start, ll_vb))


print("Start testing cgs_p model")
start = time()
ll_cgs_p = test_perplexity(cgs_p_model, phi_cgs_p)
end = time()
print("Done in {:.2f} seconds, test perplexity = {:.2f}".format(end - start, ll_cgs_p))

print("Start testing parallelized + optimized cgs model")
start = time()
ll_cgs_p = test_perplexity(parallel_cgs_model, phi_parallel_cgs, n_workers=4)
end = time()
print("Done in {:.2f} seconds, test perplexity = {:.2f}".format(end - start, ll_cgs_p))



"""Backend C++ inplementation for lda11."""
from __future__ import annotations
import typing
import numpy
import numpy.typing as npt
import scipy.sparse

_Shape = typing.Tuple[int, ...]

__all__ = [
    "LDATrainer",
    "LabelledLDATrainer",
    "Predictor",
    "learn_dirichlet",
    "learn_dirichlet_symmetric",
    "log_likelihood_doc_topic",
    "train_test_split",
]

class LDATrainer:
    def __init__(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.uint64],
        arg3: npt.NDArray[numpy.uint64],
        arg4: int,
        arg5: int,
        arg6: int,
    ) -> None: ...
    def initialize(
        self,
        arg0: npt.NDArray[numpy.int32],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
    ) -> None: ...
    def iterate_gibbs(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
        arg3: npt.NDArray[numpy.int32],
    ) -> None: ...
    def log_likelihood(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
    ) -> float: ...
    def obtain_phi(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
        arg3: npt.NDArray[numpy.int32],
    ) -> npt.NDArray[numpy.float64]: ...
    def set_doc_topic_prior(self, arg0: npt.NDArray[numpy.float64]) -> None: ...
    pass

class LabelledLDATrainer:
    def __init__(
        self,
        arg0: float,
        arg1: float,
        arg2: scipy.sparse.csr_matrix[numpy.int32],
        arg3: npt.NDArray[numpy.int32],
        arg4: npt.NDArray[numpy.uint64],
        arg5: npt.NDArray[numpy.uint64],
        arg6: int,
        arg7: int,
        arg8: int,
    ) -> None: ...
    def initialize(
        self,
        arg0: npt.NDArray[numpy.int32],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
    ) -> None: ...
    def iterate_gibbs(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
        arg3: npt.NDArray[numpy.int32],
    ) -> None: ...
    def log_likelihood(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
    ) -> float: ...
    def obtain_phi(
        self,
        arg0: npt.NDArray[numpy.float64],
        arg1: npt.NDArray[numpy.int32],
        arg2: npt.NDArray[numpy.int32],
        arg3: npt.NDArray[numpy.int32],
    ) -> npt.NDArray[numpy.float64]: ...
    pass

class Predictor:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self, arg0: int, arg1: npt.NDArray[numpy.float64], arg2: int
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def add_beta(self, arg0: npt.NDArray[numpy.float64]) -> None: ...
    def predict_gibbs(
        self,
        arg0: typing.List[npt.NDArray[numpy.int32]],
        arg1: typing.List[npt.NDArray[numpy.int32]],
        arg2: int,
        arg3: int,
        arg4: int,
        arg5: bool,
    ) -> npt.NDArray[numpy.float64]: ...
    def predict_gibbs_batch(
        self,
        arg0: typing.List[scipy.sparse.csr_matrix[numpy.int32]],
        arg1: int,
        arg2: int,
        arg3: int,
        arg4: bool,
        arg5: int,
    ) -> npt.NDArray[numpy.float64]: ...
    def predict_gibbs_with_word_assignment(
        self,
        arg0: typing.List[npt.NDArray[numpy.int32]],
        arg1: typing.List[npt.NDArray[numpy.int32]],
        arg2: int,
        arg3: int,
        arg4: int,
        arg5: bool,
    ) -> typing.Tuple[
        npt.NDArray[numpy.float64],
        typing.List[typing.Dict[int, npt.NDArray[numpy.int32]]],
    ]: ...
    def predict_mf_batch(
        self,
        arg0: typing.List[scipy.sparse.csr_matrix[numpy.int32]],
        arg1: int,
        arg2: float,
        arg3: int,
    ) -> npt.NDArray[numpy.float64]: ...
    @property
    def phis(self) -> typing.List[npt.NDArray[numpy.float64]]:
        """
        :type: typing.List[npt.NDArray[numpy.float64]]
        """
    pass

def learn_dirichlet(
    arg0: npt.NDArray[numpy.int32],
    arg1: npt.NDArray[numpy.float64],
    arg2: float,
    arg3: float,
    arg4: int,
) -> npt.NDArray[numpy.float64]:
    pass

def learn_dirichlet_symmetric(
    arg0: npt.NDArray[numpy.int32],
    arg1: float,
    arg2: float,
    arg3: float,
    arg4: int,
) -> float:
    pass

def log_likelihood_doc_topic(
    arg0: npt.NDArray[numpy.float64],
    arg1: npt.NDArray[numpy.int32],
    arg2: npt.NDArray[numpy.int32],
) -> float:
    pass

def train_test_split(
    arg0: scipy.sparse.csr_matrix[numpy.int32], arg1: float, arg2: int
) -> typing.Tuple[
    scipy.sparse.csr_matrix[numpy.int32], scipy.sparse.csr_matrix[numpy.int32]
]:
    pass



import pickle
import sys
from tempfile import NamedTemporaryFile
from typing import Tuple

import numpy as np
import numpy.typing as npt

from lda11 import LabelledLDA, labelled_lda


class LabelledLanguage:
    def __init__(
        self, TOPIC1: npt.NDArray[np.float64], TOPIC2: npt.NDArray[np.float64]
    ):
        self.topic_1: npt.NDArray[np.float64] = TOPIC1 / TOPIC1.sum()
        self.topic_2: npt.NDArray[np.float64] = TOPIC2 / TOPIC2.sum()
        self.common = np.ones_like(TOPIC1) / TOPIC1.shape[0]

    def gen_doc(
        self, n_docs: int
    ) -> Tuple[npt.NDArray[np.int64], npt.NDArray[np.int64]]:
        rns = np.random.RandomState(0)
        Xs = []
        labels = []
        for i in range(n_docs):
            cnt = rns.poisson(10)
            label = np.asfarray([1, rns.binomial(1, 0.5), rns.binomial(1, 0.5)])
            p = (
                label[0] * self.common
                + label[1] * self.topic_1
                + label[2] * self.topic_2
            )
            words = rns.multinomial(cnt, p / p.sum())
            Xs.append(words)
            labels.append(label)
        return np.vstack(Xs), np.vstack(labels)


def test_llda() -> None:
    TOPIC_A = np.asfarray([0.01, 1, 0.01, 1])
    TOPIC_B = np.asfarray([1, 0.01, 1, 0.01])
    A_word_index = np.where(TOPIC_A > 0.1)[0]
    B_word_index = np.where(TOPIC_A < 0.1)[0]

    for A_index, cgs_p in zip([1, 2], [False, True]):
        if A_index == 1:
            language = LabelledLanguage(TOPIC_A, TOPIC_B)
            B_index = 2
        else:
            language = LabelledLanguage(TOPIC_B, TOPIC_A)
            B_index = 1

        X, Y = language.gen_doc(1000)

        llda = LabelledLDA(use_cgs_p=cgs_p, n_workers=1, n_iter=100).fit(X, Y)
        if sys.platform.startswith("linux"):
            with NamedTemporaryFile() as temp_fs:
                pickle.dump(llda, temp_fs)
                temp_fs.seek(0)
                del llda
                llda_new: LabelledLDA = pickle.load(temp_fs)
        else:
            llda_new = llda

        for a_word in A_word_index:
            for b_word in B_word_index:
                assert llda_new.phi[a_word, A_index] > llda_new.phi[b_word, A_index]
                assert llda_new.phi[a_word, B_index] < llda_new.phi[b_word, B_index]

        A_DOC = np.asarray(([0, 10, 0, 10]), dtype=np.int32)
        for mode in ["mf", "gibbs"]:
            theta = llda_new.transform(A_DOC, mode=mode, n_workers=1)[0]  # type: ignore
            if A_index == 1:
                assert (theta[1] / theta[2]) > 5
            else:
                assert (theta[2] / theta[1]) > 5



from typing import List, Tuple

import numpy as np
import numpy.typing as npt

N_DOCS = 1000


class Language:
    def __init__(
        self, TOPIC1: npt.NDArray[np.float64], TOPIC2: npt.NDArray[np.float64]
    ):
        self.topic_1: npt.NDArray[np.float64] = TOPIC1 / TOPIC1.sum()
        self.topic_2: npt.NDArray[np.float64] = TOPIC2 / TOPIC2.sum()


class Docs:
    def __init__(self, languages: List[Language]):
        self.languages = languages

    def gen_doc(
        self, n_docs: int
    ) -> Tuple[List[npt.NDArray[np.int32]], npt.NDArray[np.float64]]:
        rns = np.random.RandomState(0)
        words: List[List[npt.NDArray[np.int64]]] = [
            [] for _ in range(len(self.languages))
        ]
        thetas: List[np.ndarray] = []
        for _ in range(n_docs):
            theta = rns.dirichlet(np.asfarray([0.2, 0.2]))
            thetas.append(theta)
            for lind, language in enumerate(self.languages):
                cnt = rns.poisson(5)
                wdist = (
                    float(theta[0]) * language.topic_1
                    + float(theta[1]) * language.topic_2
                )
                words[lind].append(rns.multinomial(cnt, wdist))

        return [np.vstack(x) for x in words], np.vstack(thetas)



import numpy as np
from scipy import sparse as sps

from lda11 import rowwise_train_test_split

from .conftest import Docs


def test_split(docs_gen: Docs) -> None:
    (X1, X2), _ = docs_gen.gen_doc(1000)
    X2_sp = sps.lil_matrix(X2)
    X1_tr, X1_te = rowwise_train_test_split(X1)
    assert np.all(np.asarray(X1 - X1_tr - X1_te) == 0)
    X2_tr, X2_te = rowwise_train_test_split(X2_sp, random_seed=0)
    # raise RuntimeError((X2.tocsr() - X2_tr - X2_te))
    v = np.abs(X2 - X2_tr.toarray() - X2_te.toarray()).sum()
    assert v == 0



import numpy as np
from scipy import sparse as sps

from lda11 import MultilingualLDA

from .conftest import Docs


def test_mlda(docs_gen: Docs) -> None:
    (X1, X2), true_theta = docs_gen.gen_doc(1000)
    X2 = sps.lil_matrix(X2)
    lda = MultilingualLDA(
        2, n_iter=50, optimize_interval=1, optimize_burn_in=25, use_cgs_p=False
    )
    lda.fit(X1, X2)
    phi1, phi2 = lda.phis

    # determin which is TOPIC1

    lang1_topic1_strong_index = np.where(docs_gen.languages[0].topic_1 > 0.1)[0]
    lang1_topic2_strong_index = np.where(docs_gen.languages[0].topic_1 < 0.1)[0]
    if (
        phi1[lang1_topic1_strong_index, 0].mean()
        > phi1[lang1_topic2_strong_index, 0].mean()
    ):
        topic1_index = 0
        topic2_index = 1
    else:
        topic1_index = 1
        topic2_index = 0
    for i in lang1_topic1_strong_index:
        for j in lang1_topic2_strong_index:
            assert phi1[i, topic1_index] > phi1[j, topic1_index]
            assert phi1[i, topic2_index] < phi1[j, topic2_index]

    lang2_topic1_strong_index = np.where(docs_gen.languages[1].topic_1 > 0.1)[0]
    lang2_topic2_strong_index = np.where(docs_gen.languages[1].topic_1 < 0.1)[0]
    for i in lang2_topic1_strong_index:
        for j in lang2_topic2_strong_index:
            assert phi2[i, topic1_index] > phi2[j, topic1_index]
            assert phi2[i, topic2_index] < phi2[j, topic2_index]

    # just check it works.
    for algo in ["mf", "gibbs"]:
        checked_cnt = 0
        theta_inferred = lda.transform(X1, X2, mode=algo)  # type: ignore
        for i in range(X1.shape[0]):
            if (true_theta[i, 0] / true_theta[i, 1]) > 10 and (X1[i].sum() > 5):
                checked_cnt += 1
                assert theta_inferred[i, topic1_index] > theta_inferred[i, topic2_index]
        assert checked_cnt > 0

    wdt = lda.word_topic_assignment(X1, X2)
    assert len(wdt) == 1000
    for i, wdt_result_doc in enumerate(wdt):
        theta = wdt_result_doc[0]
        if (true_theta[i, 0] / true_theta[i, 1]) > 10:
            assert theta[topic1_index] > theta[topic2_index]
        m = wdt_result_doc[1]
        assert len(m) == 2
        # lang 1
        lang1_assignment = m[0]
        for word, topic in lang1_assignment.items():
            if (topic[topic1_index] / (1e-10 + topic[topic2_index])) > 10:
                assert word in lang1_topic1_strong_index
            if (topic[topic2_index] / (1e-10 + topic[topic1_index])) > 10:
                assert word in lang1_topic2_strong_index

        lang2_assignment = m[1]
        for word, topic in lang2_assignment.items():
            if (topic[topic1_index] / (1e-10 + topic[topic2_index])) > 10:
                assert word in lang2_topic1_strong_index
            if (topic[topic2_index] / (1e-10 + topic[topic1_index])) > 10:
                assert word in lang2_topic2_strong_index



from typing import Optional, Tuple

import numpy as np
from scipy import sparse as sps

from ._lda import train_test_split
from .lda import IntegerType, RealType, ValidXType


def rowwise_train_test_split(
    X: ValidXType, random_seed: Optional[int] = None, test_ratio: float = 0.5
) -> Tuple[sps.csr_matrix, sps.csr_matrix]:
    """
    split matrix randomly
    """
    if random_seed is None:
        random_seed = np.random.randint(-(2 ** 31), 2 ** 31 - 1)
    X = sps.csr_matrix(X, dtype=IntegerType)
    return train_test_split(X, test_ratio, random_seed)



from typing import Optional

import numpy as np
from numpy import typing as npt
from scipy import sparse as sps
from tqdm import tqdm

from ._lda import LabelledLDATrainer
from ._lda import Predictor as CorePredictor
from .lda import (
    IntegerType,
    LDAPredictorMixin,
    RealType,
    ValidXType,
    check_array,
    number_to_array,
)


class LabelledLDA(LDAPredictorMixin):
    def __init__(
        self,
        alpha: float = 1e-2,
        epsilon: float = 1e-30,
        n_iter: int = 1000,
        n_workers: int = 1,
        use_cgs_p: bool = True,
    ):
        self.n_components_: Optional[int] = None
        self.alpha = alpha
        self.epsilon = 1e-20
        self.n_vocabs = None
        self.docstate_ = None
        self.components_: Optional[npt.NDArray[np.int32]] = None
        self.predictor = None
        self.n_workers = n_workers
        self.epsilon = epsilon
        self.n_iter = n_iter
        self.use_cgs_p = use_cgs_p

    def fit(self, X: ValidXType, Y: ValidXType) -> "LabelledLDA":
        self._fit_llda(X, Y)
        return self

    def _fit_llda(
        self,
        X: ValidXType,
        Y: ValidXType,
    ) -> npt.NDArray[np.int32]:
        if not sps.issparse(Y):
            Y = sps.csr_matrix(Y).astype(IntegerType)
        else:
            Y = Y.astype(IntegerType)

        self.n_components = int(Y.shape[1])
        self.topic_word_priors_ = [
            number_to_array(X.shape[1], 1 / float(self.n_components), None)
        ]

        try:
            count, dix, wix = check_array(X)
        except:
            print("Check for X failed.")
            raise

        doc_topic = np.zeros((X.shape[0], self.n_components), dtype=IntegerType)
        word_topic = np.zeros((X.shape[1], self.n_components), dtype=IntegerType)
        topic_counts = np.zeros(self.n_components, dtype=IntegerType)

        docstate = LabelledLDATrainer(
            self.alpha,
            self.epsilon,
            Y,
            count,
            dix,
            wix,
            self.n_components,
            42,
            self.n_workers,
        )
        docstate.initialize(word_topic, doc_topic, topic_counts)

        with tqdm(range(self.n_iter)) as pbar:
            for _ in pbar:
                docstate.iterate_gibbs(
                    self.topic_word_priors_[0], doc_topic, word_topic, topic_counts
                )

        doc_topic_prior = self.alpha * np.ones(self.n_components, dtype=RealType)

        self.components_ = word_topic.transpose()
        predictor = CorePredictor(self.n_components, doc_topic_prior, 42)
        if self.use_cgs_p:
            phi = docstate.obtain_phi(
                self.topic_word_priors_[0], doc_topic, word_topic, topic_counts
            )
        else:
            phi = word_topic + self.topic_word_priors_[0][:, np.newaxis]
            phi /= phi.sum(axis=0)[np.newaxis, :]
            phi = phi.transpose()
        predictor.add_beta(phi.transpose())
        self.predictor = predictor

        return doc_topic

    @property
    def phi(self) -> npt.NDArray[np.float64]:
        assert self.predictor is not None
        return self.predictor.phis[0]
